import pip
import sys, os
import emucorebrain.consts.consts as consts
import json
from emucorebrain.data.models.route_model import RouteModel
import importlib


# Installs the dependencies required.
def _install_dependencies(list_deps : list):
    for dep in list_deps:
        args = ["-q"]
        pip.main(["install"] + args + [dep])


# Checks whether the given class instance is a valid instance of the class Process.
def _is_valid_process(class_object):
    try:
        return class_object.VALID_PROCESS

    except AttributeError:
        return False


# TODO: CODE THERE TO SUIT PROCESS INTERFACE
# Installs dependencies, Checks whether the given namespace directory for processes is valid.
def _run_validation_on_namespace_dir(processes_namespaces_folder_path):
    structure_definitions_file_path = processes_namespaces_folder_path + "/" + consts.PROCESSES_STRUCT_FILE_FILENAME
    if os.path.exists(structure_definitions_file_path):
        deps_file_path = processes_namespaces_folder_path + "/" + consts.PROCESSES_DEPS_FILE_FILENAME
        if os.path.exists(deps_file_path):
            deps_file = open(deps_file_path, "rb")
            deps_data = json.load(deps_file)

            list_dep_data = []
            for dep_index in deps_data:
                dep_index_data = deps_data[dep_index]

                if consts.PROCESSES_DEPS_FILE_PROP_DEP_NAME in dep_index_data and consts.PROCESSES_DEPS_FILE_PROP_DEP_VERSION in dep_index_data:
                    dep_index_name = dep_index_data[consts.PROCESSES_DEPS_FILE_PROP_DEP_NAME]
                    dep_index_version = dep_index_data[consts.PROCESSES_DEPS_FILE_PROP_DEP_VERSION]
                    dep_index_req_line = dep_index_name + "==" + dep_index_version

                    list_dep_data.append(dep_index_req_line)
                else:
                    raise Exception("Dependencies file contains malformed(missing name/version) near the dependency index: " + dep_index)

            _install_dependencies(list_dep_data)

            structure_definitions_file_data = open(structure_definitions_file_path).read()
            structure_definitions_file_data = json.loads(structure_definitions_file_data)

            if all(basic_prop in structure_definitions_file_data for basic_prop in consts.PROCESSES_STRUCT_FILE_BASIC_PROPERTY_KEYS):
                sys.path.append(os.path.abspath(processes_namespaces_folder_path))

                processes = structure_definitions_file_data[consts.PROCESSES_STRUCT_FILE_PROP_PROCESSES]
                for process in processes:
                    process_folder = process[consts.PROCESSES_STRUCT_FILE_PROP_PROCESSES_NAMESPACE]
                    process_folder_path = processes_namespaces_folder_path + "/" + process_folder

                    if os.path.exists(process_folder_path):
                        process_route_model = RouteModel(process[consts.TASKS_STRUCT_FILE_PROP_EXECUTORS_CLASS])
                        process_name = process_route_model.get_class_name()
                        process_file_name = process_name + ".py"
                        process_file_path = process_folder_path + "/" + process_file_name

                        if not os.path.exists(process_file_path):
                            raise Exception("Class File: " + process_name + " cannot be found inside the Namespace Folder: " + process_folder + ".")
                        else:
                            # Each namespace(task) file should have its class with the same name as the file
                            try:
                                sys.path.append(os.path.abspath(process_folder_path))

                                unique_class_import_name = process_name + "." + process_name
                                unique_class_object = getattr(importlib.import_module(unique_class_import_name), process_name)
                                assert _is_valid_process(unique_class_object)

                            except AssertionError:
                                raise Exception("One of the namespaces in the tasks directory does not implement Process.")

                            except Exception:
                                raise Exception("Unexpected error. MAYBE the file(s) in the tasks directory you've provided contains ambiguous class names.")
                    else:
                        raise Exception("Namespace folder: " + process_folder + " not found.")

            else:
                raise Exception("One or more of the following required entries are not found in the " + consts.PROCESSES_STRUCT_FILE_FILENAME + ".\n" + consts.PROCESSES_STRUCT_FILE_PROP_DEP_DIRS + ", " + consts.PROCESSES_STRUCT_FILE_PROP_PROCESSES)
        else:
            raise Exception("Namespace folder: " + processes_namespaces_folder_path + " does not contain the dependencies folder: " + consts.PROCESSES_DEPS_FILE_FILENAME)
    else:
        raise Exception(consts.PROCESSES_STRUCT_FILE_FILENAME + " does not exist inside " + processes_namespaces_folder_path + ".\nPlease make sure the processes folder path you've given contains valid implementations of Process interface.")


def get_loaded_namespaces(procceses_namespaces_folderpath):
    # If the validation occurs successfully, we get the sys.path appended with all the namespace directories and
    # the parent directory. Therefore it in not necessary to append them here again.
    _run_validation_on_namespace_dir(procceses_namespaces_folderpath)

    class_namespaces = []

    structure_definitions_file_path = procceses_namespaces_folderpath + "/" + consts.PROCESSES_STRUCT_FILE_FILENAME
    structure_definitions_file_data = open(structure_definitions_file_path).read()
    structure_definitions_file_data = json.loads(structure_definitions_file_data)
    processes = structure_definitions_file_data[consts.PROCESSES_STRUCT_FILE_PROP_PROCESSES]
    for process in processes:
        process_route_model = RouteModel(process[consts.PROCESSES_STRUCT_FILE_PROP_PROCESSES_CLASS])
        process_name = process_route_model.get_class_name()

        unique_class_import_name = process_name + "." + process_name
        class_namespace_class = getattr(importlib.import_module(unique_class_import_name), process_name)
        class_namespace_instance = class_namespace_class()
        class_namespaces.append(class_namespace_instance)

    return class_namespaces



import importlib


class RouteModel:

    INDEX_CLASS = 0
    INDEX_METHOD = 1

    METHOD_SEPARATOR = '.'
    DEFAULT_NAME_RUN_METHOD = "run"

    def __init__(self, route_str):
        self._route_str: str = route_str

    def get_class_name(self):
        route_sep = self._route_str.split(self.METHOD_SEPARATOR)
        return route_sep[self.INDEX_CLASS]

    def get_method_name(self):
        route_sep = self._route_str.split(self.METHOD_SEPARATOR)
        if len(route_sep) > self.INDEX_METHOD:
            return route_sep[self.INDEX_METHOD]
        else:
            return RouteModel.DEFAULT_NAME_RUN_METHOD

    # Please note that all the paths to the Class namespaces should be added prior to calling this method.
    def get_class_instance(self):
        class_name = self.get_method_name()
        unique_class_import_name = class_name + "." + class_name
        class_namespace_class = getattr(importlib.import_module(unique_class_import_name), class_name)
        class_namespace_instance = class_namespace_class()
        return class_namespace_instance

    def get_method(self):
        class_namespace_instance = self.get_class_instance()
        name_method = self.get_method_name()
        return getattr(class_namespace_instance, name_method)

    def get_method_by_class_instance(self, class_instance):
        name_method = self.get_method_name()
        return getattr(class_instance, name_method)



import abc

# This is the "interface" class for all the processes.
# Each an every process defined, should extend this class.


# Processes are executed throughout the execution life cycle, regardless of user's input.
# Therefore must use lockers together with them.
class Process(abc.ABC):

    VALID_PROCESS = True

    # This method is used to start the process.
    # Usually called at initializing function.
    # Can be used in any form can initialize variables/can actually start the process or any other stuff.
    # Note: Never use threads to run the iteration method, run it over the exec_iter method to ensure that I/O doesn't get mixed up.
    @abc.abstractmethod
    def start_process(self, args):
        pass

    # This method is executed in each cycle of the main loop in main program.
    # The code that goes in depends on the process that is to be executed.
    @abc.abstractmethod
    def exec_iter(self):
        pass

    # Starts/Resumes a process that has been stopped/paused by pause_process method.
    # The code that goes in depends on the process that is to be executed.
    @abc.abstractmethod
    def resume_process(self):
        pass

    # This method is used to temporarily stop(pause) the process which can be started again using resume_process method
    # The code that goes in depends on the process that is to be executed.
    @abc.abstractmethod
    def pause_process(self):
        pass

    # This method is executed when the system wraps up or the process is needed to be stopped.
    # The code that goes in depends on the process.
    # Usually clean-up code goes in here.
    @abc.abstractmethod
    def destroy_process(self):
        pass



from emucorebrain.data.models.lockers import LockerTypes


class LockersContainer:

    def __init__(self):
        self._lockers = {}
        self._locker_index = 0

    def get_lockers(self):
        return self._lockers

    def is_inputs_locked(self):
        for _, locker_type in self._lockers.items():
            if locker_type == LockerTypes.INPUT_MECHANISMS:
                return True

        return False

    def is_outputs_locked(self):
        for _, locker_type in self._lockers.items():
            if locker_type == LockerTypes.OUTPUT_MECHANISMS:
                return True

        return False

    def add_locker(self, locker_type):
        self._lockers[self._locker_index] = locker_type
        self._locker_index += 1
        return self._locker_index - 1

    def remove_locker(self, locker_index):
        del self._lockers[locker_index]
